/**
 * Copyright (c) 2008 Washington University
 */
package org.nrg.xnat.engine;

import java.io.File;
import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;
import org.hsqldb.jdbc.jdbcDataSource;
import org.mortbay.jetty.Server;
import org.mortbay.jetty.servlet.ServletHolder;
import org.nrg.xnat.filestore.FileStore;
import org.nrg.xnat.filestore.FileStoreServlet;
import org.nrg.xnat.filestore.URIPathFileStore;
import org.nrg.xnat.search.SearchServlet;
import org.nrg.xnat.tags.TagLabelsServlet;

/**
 * @author Kevin A. Archie <karchie@npg.wustl.edu>
 *
 */
public final class Engine {
    private final static String SERVLET_CONTEXT_TAGS = "/tags";
    private final static String SERVLET_CONTEXT_STORE = "/data";
    private final static String SERVLET_CONTEXT_SEARCH = "/search";

    // Used to convert local relative URIs to a relative URI that includes servlet context.
    // Must be consistent with SERVLET_CONTEXT_STORE above.
    private final static String CONTEXT_STORE_PREFIX = "data/";

    private final static String dbUser = "sa";
    private final static String dbPassword = "";

    private final Logger log = Logger.getLogger(Engine.class);
    private final jdbcDataSource hsqlds;
    private Server server = null;

    private Engine() { throw new UnsupportedOperationException(); }

    /**
     * Start a fileserver database
     * @param dbFile Database file
     * @param storeRoot Root directory for resource storage
     * @throws Exception
     */
    private Engine(final File dbFile, final File storeRoot) throws SQLException,IOException {
	// Set up the database and filestore
	hsqlds = new jdbcDataSource();
	hsqlds.setDatabase("jdbc:hsqldb:file:" + dbFile.getPath().replace(File.separatorChar, '/'));
	hsqlds.setUser(dbUser);
	hsqlds.setPassword(dbPassword);
	
	// Set up the database tables before creating any objects that use the database
	final Connection c = hsqlds.getConnection();
	try {
	    final Statement s = c.createStatement();
	    try {
		s.execute("CREATE MEMORY TABLE tag_labels ( tag_label_id INTEGER IDENTITY, tag_label VARCHAR, tag_label_lc VARCHAR, UNIQUE(tag_label_lc) )");
	    } catch (SQLException e) { log.debug(e); }
	    try {
		s.execute("CREATE MEMORY TABLE mime_media_types ( mime_media_type_id INTEGER IDENTITY, mime_media_type VARCHAR, UNIQUE(mime_media_type) )");
	    } catch (SQLException e) { log.debug(e); }
	    try {
		s.execute("CREATE CACHED TABLE resources ( resource_id INTEGER IDENTITY, uri VARCHAR, resource VARCHAR, mime_media_type_id INTEGER, "
			+ "UNIQUE(uri), CONSTRAINT fk_mime_media_type_id FOREIGN KEY (mime_media_type_id) REFERENCES mime_media_types (mime_media_type_id))");
	    } catch (SQLException e) { log.debug(e); }
	    try {
		s.execute("CREATE CACHED TABLE resource_tags ( resource_id INTEGER, tag_label_id INTEGER, value VARCHAR, "
			+ "UNIQUE(resource_id, tag_label_id, value), "
			+ "CONSTRAINT fk_resource_id FOREIGN KEY (resource_id) REFERENCES resources (resource_id), "
			+ "CONSTRAINT fk_tag_label_id FOREIGN KEY (tag_label_id) REFERENCES tag_labels (tag_label_id))");
	    } catch (SQLException e) { log.debug(e); }
	    
	    try {
		s.execute("CREATE INDEX tag_label_id_idx ON resource_tags ( tag_label_id, value )");
	    } catch (SQLException e) { log.debug(e); }
	    
	    try {
		s.execute("CREATE TEMP TABLE search_matches ( resource_id INTEGER, uri VARCHAR, resource VARCHAR )");
	    } catch (SQLException e) { log.debug(e); }
	    try {
		s.execute("CREATE TEMP TABLE search_match_tags ( uri VARCHAR, tag_label_id INTEGER, value VARCHAR )");
	    } catch (SQLException e) { log.debug(e); }

	    try { 
		s.execute("SET CHECKPOINT DEFRAG 10");	// CHECKPOINT does DEFRAG if .data file > 10 MB
	    } catch (SQLException e) { log.debug(e); }
	    
	    try {
		s.execute("CHECKPOINT");
	    } catch (SQLException e) { log.error(e); }
	    s.close();
	} finally {
	    c.close();
	}

	
	final FileStore store = new URIPathFileStore(SERVLET_CONTEXT_STORE, storeRoot);

	// Put database info into JNDI
	try {
	    final Context context = new InitialContext();
	    final Context compContext = (Context)context.lookup("java:comp");
	    final Context envContext = compContext.createSubcontext("env");
	    final Context jdbcContext = envContext.createSubcontext("jdbc");
	    jdbcContext.bind("datasource", hsqlds);

	    final Context xnatContext = envContext.createSubcontext("xnat");
	    xnatContext.bind("store", store);
	    xnatContext.bind("storeContext", CONTEXT_STORE_PREFIX);
	} catch (NamingException e) {
	    // This only occurs if there's been a programming error above, I think.
	    throw new RuntimeException("error in JNDI configuration", e);
	}

    }

    
    /**
     * Set the maximum size for the HSQLDB .log file; when the .log file reaches
     * this size, the database will shut down briefly to commit changes, then
     * restart with a cleared .log file.
     * @param size maximum size of HSQLDB .log file, in MB
     * @throws SQLException
     */
    public void setLogsize(final int size) throws SQLException {
	final Connection c = hsqlds.getConnection();
	try {
	    final Statement s = c.createStatement();
	    try {
	        s.execute(String.format("SET LOGSIZE %d", size));
	    } finally {
	        s.close();
	    }
	} finally {
	    c.close();
	}
    }
    
    
    /**
     * Starts the web services on the given port.
     * @param port TCP port for web services
     * @param block true if call should block, false if it should return.
     */
    public void startWebServices(final int port, boolean block) {
	if (null != server) {
	    throw new IllegalStateException("web services server " + server + " already running");
	}
	server = new Server(port);

	// Set up servlets and contexts
	final org.mortbay.jetty.servlet.Context rootContext = new org.mortbay.jetty.servlet.Context(server, "/");
	rootContext.addServlet(new ServletHolder(new TagLabelsServlet()), SERVLET_CONTEXT_TAGS);
	rootContext.addServlet(new ServletHolder(new FileStoreServlet()), SERVLET_CONTEXT_STORE + "/*");
	rootContext.addServlet(new ServletHolder(new SearchServlet()), SERVLET_CONTEXT_SEARCH);

	try {
	    server.start();
	    if (block) {
		server.join();
	    }
	} catch (Exception e) {
	    log.error(e);
	}
    }


    public void stopWebServices() {
	if (null == server) {
	    throw new IllegalStateException("web services server not running");
	}
	try {
	    server.stop();
	} catch (Exception e) {
	    log.error(e);
	} finally {
	    server = null;
	}
    }

    
    public void dispose() {
	engine = null;
	stopWebServices();
	final Connection c;
	try {
	    c = hsqlds.getConnection();
	} catch (SQLException e) {
	    log.error("Can't obtain DB connection for shutdown", e);
	    return;
	}
	try {
	    final Statement s = c.createStatement();
	    try {
		s.execute("SHUTDOWN COMPACT");
	    } finally {
		s.close();
	    }
	} catch (SQLException e) {
	    log.error("SQL error shutting down", e);
	} finally {
	    try { c.close(); } catch (SQLException ignore) {}
	}
    }

    
    public void checkpoint() {
	try {
	    final Connection c = hsqlds.getConnection();
	    try {
		final Statement s = c.createStatement();
		try {
		    s.execute("CHECKPOINT");
		} finally {
		    s.close();
		}
	    } finally {
		c.close();
	    }
	} catch (SQLException e) {
	    log.error("Unable to checkpoint database", e);
	}
    }
    
    
    private static Engine engine = null;

    /**
     * Returns the XNAT engine (singleton) object, initializing it if it doesn't already exist.
     * @param dbFile File path for underlying HSQLDB database
     * @param storeRoot File path of root directory for items stored via web services
     * @return XNAT engine instance
     * @throws SQLException
     * @throws IOException
     */
    public static Engine getEngine(final File dbFile, final File storeRoot) throws SQLException, IOException {
	if (null == engine) {
	    engine = new Engine(dbFile, storeRoot);
	}
	return engine;
    }

    /**
     * Returns the XNAT engine (singleton) object, which must already be initialized.
     * @return XNAT engine instance
     */
    public static Engine getEngine() {
	if (null == engine)
	    throw new IllegalStateException("XNAT engine not initialized");
	return engine;
    }


    /**
     * @param args
     */
    public static void main(String[] args) throws Exception {
	if (args.length < 3) {
	    System.err.println("Usage: Engine <port> <db-path> <store-path>");
	    return;
	}
	BasicConfigurator.configure();
	final Engine engine = new Engine(new File(args[1]), new File(args[2]));
	engine.startWebServices(Integer.valueOf(args[0]), true);
    }

}
