/**
 * Copyright (c) 2008 Washington University
 */
package org.nrg.xnat.search;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang.StringEscapeUtils;
import org.apache.log4j.Logger;
import org.nrg.xnat.sql.AbstractResultIterator;
import org.nrg.xnat.tags.InvalidTagLabelException;
import org.nrg.xnat.tags.TagLabel;
import org.nrg.xnat.tags.TagLabelException;
import org.nrg.xnat.tags.UndefinedTagLabelException;
import org.nrg.xnat.util.FinalMapEntry;
import org.nrg.xnat.util.KeyIterator;


/**
 * @author Kevin A. Archie <karchie@npg.wustl.edu>
 *
 */
public final class Search {
    private static final class ResourceEntry extends FinalMapEntry<URI,File> {
	private ResourceEntry(final String s_uri, final String path) throws URISyntaxException {
	    super(new URI(s_uri), null == path ? null : new File(path));
	}

	private static final class EntryIterator extends AbstractResultIterator<Map.Entry<URI,File>> {
	    private EntryIterator(final ResultSet rs, final boolean shouldCloseConnection) {
		super(rs, shouldCloseConnection);
	    }

	    private EntryIterator(final ResultSet rs) { super(rs, true); }

	    public Map.Entry<URI,File> makeEntry(final ResultSet rs) throws SQLException {
		try {
		    return new ResourceEntry(rs.getString(1), rs.getString(2));
		} catch (URISyntaxException e) {
		    throw new RuntimeException("Bad database contents: invalid resource uri", e);
		}
	    }
	}
    }


    private final Logger logger = Logger.getLogger(Search.class);
    private final Map<TagLabel,Collection<String>> tagValues;	// must be collected
    private final ResourceEntry.EntryIterator resourceEntryIterator;

    /**
     * Creates a new Search object from the given ResultSet.  The Search object is
     * responsible for freeing both Connection and ResultSet resources.
     * @param c Connection for which the Search object is responsible (if non-null)
     * @param rs ResultSet for the search results
     */
    private Search(final ResultSet rs, final Map<TagLabel,Collection<String>> metadata) {
	resourceEntryIterator = new ResourceEntry.EntryIterator(rs, true);
	this.tagValues = metadata;
    }

    public Iterator<Map.Entry<URI,File>> resourceEntryIterator() {
	return resourceEntryIterator;
    }

    public Iterator<URI> resourceIterator() {
	return new KeyIterator<URI,File>(resourceEntryIterator);
    }

    public Collection<TagLabel> getTagLabels() {
	return new ArrayList<TagLabel>(tagValues.keySet());
    }

    public Collection<String> getTagValues(final TagLabel tag) {
	final Collection<String> c = tagValues.get(tag);
	return null == c ? new ArrayList<String>() : new ArrayList<String>(c);
    }

    public void dispose() { resourceEntryIterator.dispose(); }


    private static interface SearchParam {
	public String getLabel();
	public String getQueryPart();
    }


    private final static class SimpleSearchParam implements SearchParam {
	private final TagLabel label;
	private final String value;

	SimpleSearchParam(final TagLabel label, final String value) {
	    this.label = label;
	    this.value = StringEscapeUtils.escapeSql(value);
	}

	SimpleSearchParam(final String label, final String value) throws InvalidTagLabelException {
	    this(new TagLabel(label), value);
	}

	public final String getLabel() { return label.toString(); }

	public final String toString() { return String.format("%s=\'%s\'", label, value); }

	public final String getQueryPart() {
	    return String.format("SELECT rt.resource_id FROM tag_labels tl, resource_tags rt WHERE rt.value = '%s' AND rt.tag_label_id = tl.tag_label_id AND tl.tag_label_lc = LOWER('%s')",
		    value, label);
	}
    }


    private final static class IsSetSearchParam implements SearchParam {
	private final TagLabel label;

	IsSetSearchParam(final TagLabel label) { this.label = label; }

	public final String getLabel() { return label.toString(); }

	public final String toString() { return String.format("%s is set", label); }

	public final String getQueryPart() {
	    return String.format("SELECT DISTINCT rt.resource_id FROM tag_labels tl, resource_tags rt WHERE rt.tag_label_id = tl.tag_label_id AND tl.tag_label_lc = LOWER('%s')",
		    label);
	}
    }


    private static String conjoinQueries(final String conjunction, final Collection<SearchParam> params) {
	if (params.isEmpty()) {
	    throw new IllegalArgumentException("cannot conjoin zero queries");
	}
	final List<SearchParam> ps = new ArrayList<SearchParam>(params);
	final StringBuilder sb = new StringBuilder(ps.get(0).getQueryPart());
	for (int i = 1; i < ps.size(); i++) {
	    sb.append(conjunction);
	    sb.append(ps.get(i).getQueryPart());
	}
	return sb.toString();
    }


    private static String intersectQueries(final Collection<SearchParam> params) {
	return conjoinQueries(" INTERSECT ", params);
    }


    private static final String SQL_GET_TAG_ID = "SELECT tag_label_id FROM tag_labels WHERE tag_label_lc = LOWER(?)";
    private static final String SQL_GET_TAG_VALUES = "SELECT DISTINCT value FROM search_match_tags WHERE tag_label_id = ?";


    /**
     * Performs a search in the database for matching resources, based on exact
     * tag value matches and regular expression value matches, and unset tags.
     * This method is NOT thread-safe for threads that share a Connection
     * instance, in that it changes the autocommit behavior for the Connection
     * while using temporary tables.
     * 
     * @param c
     *                Connection to underlying SQL database
     * @param values
     *                Map from tag labels to constraining values, either a
     *                single value (typically String) or a Collection of values
     *                (typically String), all of which must be satisfied. Tag
     *                labels must be valid and defined in the database. The
     *                values need not be SQL-safe.
     * @param set
     *                Labels of tags that must be set (any value). Labels must
     *                be valid and defined in the database.
     * @param unset
     *                Labels of tags that must be unset. Labels must be valid
     *                and defined in the database.
     * @param limit
     *                Maximum number of resource records to be returned. If
     *                null, there is no limit.
     * @return Results object from which detailed search results can be read
     * @throws TagLabelException
     *                 if a tag label is invalid or not defined in the database
     * @throws SQLException
     *                 if an error occurs
     */
    @SuppressWarnings("unchecked")
    public static Search get(final Connection c,
	    final Map<TagLabel,Collection<String>> values, final Iterable<TagLabel> set, final Iterable<TagLabel> unset,
	    final Integer limit)
    throws TagLabelException,SQLException {
	final Logger logger = Logger.getLogger(Search.class);
	
	final Collection<SearchParam> matches = new ArrayList<SearchParam>();
	for (final Map.Entry<TagLabel,Collection<String>> me : values.entrySet()) {
	    final TagLabel tl = me.getKey();
	    for (final String value : new LinkedHashSet<String>(me.getValue())) {
		matches.add(new SimpleSearchParam(tl, value));
	    }
	}
	for (final TagLabel label : set) {
	    matches.add(new IsSetSearchParam(label));
	}

	// First look for set constraint matches
	final StringBuilder matchSQL = new StringBuilder("INSERT INTO search_matches SELECT resource_id, uri, resource FROM resources ");
	if (!matches.isEmpty()) {
	    matchSQL.append(" WHERE resource_id IN ( ");
	    matchSQL.append(intersectQueries(matches));
	    matchSQL.append(" )");
	}
	logger.debug(matchSQL.toString());

	final Map<TagLabel,Collection<String>> tagValues = new HashMap<TagLabel,Collection<String>>();

	// This statement is handed off (implicitly) to the returned Search,
	// so don't close it when we're done here.
	final Statement s = c.createStatement();
	final ResultSet searchResults;
	try {
	    // Hold the temp tables through this block.
	    c.setAutoCommit(false);

	    // Get the resources matching the set-tag constraints (into table search_matches)
	    s.executeUpdate(matchSQL.toString());

	    // Get tag label IDs for all unset tag constraints
	    final Map<TagLabel,Integer> unsetLabels = new HashMap<TagLabel,Integer>();
	    if (unset.iterator().hasNext()) {
		final PreparedStatement getTagID = c.prepareStatement(SQL_GET_TAG_ID);
		try {
		    for (final TagLabel label : unset) {
			getTagID.setString(1, label.toString());
			final ResultSet labelrs = getTagID.executeQuery();
			if (!labelrs.next()) {
			    throw new UndefinedTagLabelException(label);
			}
			try {
			    unsetLabels.put(label, labelrs.getInt(1));
			} finally {
			    labelrs.close();
			}
		    }
		} finally {
		    getTagID.close();
		}

		// We'll need tags values for both checking unset constraints and listing remaining values
		final StringBuilder tagValsSQL = new StringBuilder("INSERT INTO search_match_tags SELECT sm.uri, rt.tag_label_id, rt.value FROM search_matches sm INNER JOIN resource_tags rt ON sm.resource_id = rt.resource_id AND rt.tag_label_id IN ( ");
		join(tagValsSQL, unsetLabels.values(), ", ");
		tagValsSQL.append(" )");
		s.executeUpdate(tagValsSQL.toString());
	    }

	    // Need to include a table alias here so the (optional) LIMIT below doesn't get intrepreted as
	    // an alias definition for search_matches.
	    final StringBuilder resourcesSQL = new StringBuilder("SELECT uri, resource FROM search_matches sm");

	    // Matching resources must also have unset tags unset
	    if (!unsetLabels.isEmpty()) {
		resourcesSQL.append(" WHERE uri NOT IN ( SELECT uri FROM search_match_tags )");
	    }
	    if (null != limit) {
		resourcesSQL.append(" LIMIT ");
		resourcesSQL.append(limit);
	    }
	    
	    
	    // Get values for all specified-unset tags
	    if (!unsetLabels.isEmpty()) {
		final PreparedStatement getTagValues = c.prepareStatement(SQL_GET_TAG_VALUES);
		try {
		    for (Map.Entry<TagLabel,Integer> te : unsetLabels.entrySet()) {
			getTagValues.setInt(1, te.getValue());
			final ResultSet tvrs = getTagValues.executeQuery();
			try {
			    while (tvrs.next()) {
				final TagLabel tl = te.getKey();
				if (!tagValues.containsKey(tl)) {
				    tagValues.put(tl, new ArrayList<String>());
				}
				tagValues.get(tl).add(tvrs.getString(1));
			    }
			} finally {
			    tvrs.close();
			}
		    }
		} finally {
		    getTagValues.close();
		}
	    }
	    searchResults = s.executeQuery(resourcesSQL.toString());
	} catch (SQLException e) {
	    try { s.close(); } catch (SQLException ignore) {}
	    throw e;
	} finally {
	    try { s.executeUpdate("COMMIT"); } catch (SQLException ignore) {}
	    try {
		c.setAutoCommit(true);
	    } catch (SQLException e) {
		Logger.getLogger(Search.class).error("Failed to restore AUTOCOMMIT state", e);
	    }
	}

	return new Search(searchResults, tagValues);
    }

    public static Map<TagLabel,Collection<String>> getTags(final Connection c,
	    final Map<TagLabel,Collection<String>> constraints,
	    final Collection<TagLabel> requested) throws SQLException,TagLabelException {
	final Collection<SearchParam> matches = new ArrayList<SearchParam>();
	for (final Map.Entry<TagLabel,Collection<String>> me : constraints.entrySet()) {
	    final TagLabel tl = me.getKey();
	    for (final String value : new LinkedHashSet<String>(me.getValue())) {
		if ("*".equals(value)) {
		    throw new IllegalArgumentException("tag values must be real (no wildcards)");
		}
		matches.add(new SimpleSearchParam(tl, value));
	    }
	}

	final Statement s = c.createStatement();
	try {
	    final PreparedStatement getValues = c.prepareStatement("SELECT DISTINCT value FROM resource_tags rt, tag_labels tl WHERE rt.resource_id = ? AND rt.tag_label_id = tl.tag_label_id AND tl.tag_label_lc = LOWER(?)");
	    try {
		final ResultSet resources = s.executeQuery(intersectQueries(matches));
		try {
		    final Map<TagLabel,Collection<String>> values = new LinkedHashMap<TagLabel,Collection<String>>();
		    while (resources.next()) {
			for (final TagLabel tl : new LinkedHashSet<TagLabel>(requested)) {	// eliminate duplicates
			    getValues.setInt(1, resources.getInt(1));
			    getValues.setString(2, tl.toString());
			    final ResultSet vrs = getValues.executeQuery();
			    try {
				final Collection<String> v = new ArrayList<String>();

				while (vrs.next()) {
				    v.add(vrs.getString(1));
				}
				assert !values.containsKey(tl);
				values.put(tl, v);
			    } finally {
				try { vrs.close(); } catch (SQLException ignore) {}
			    }
			}
		    }
		    return values;
		} finally {
		    try { resources.close(); } catch (SQLException ignore) {}
		}
	    } finally {
		try { getValues.close(); } catch (SQLException ignore) {}
	    }
	} finally {
	    try { s.close(); } catch (SQLException ignore) {}
	}
    }


    public static Search getCompletions(final Connection c, final URI uri) throws SQLException {
	// Get a path that has exactly one path separator at the end.
	final String rawPath = uri.toString();
	final String path = rawPath.replaceAll("/*\\Z", "/");
	final String sql = String.format("SELECT uri, resource FROM resources WHERE uri LIKE '%s%%'", StringEscapeUtils.escapeSql(path));

	final Statement s = c.createStatement();
	try {
	    return new Search(s.executeQuery(sql), null);
	} catch (SQLException e) {
	    s.close();
	    throw e;
	}
    }


    private static void join(final StringBuilder sb, final Iterable<?> objs, final String separator) {
	final Iterator<?> i = objs.iterator();
	if (i.hasNext()) {
	    sb.append(i.next().toString()); 
	    while (i.hasNext()) {
		sb.append(separator);
		sb.append(i.next());
	    }
	}
    }
}
