/**
 * Copyright (c) 2008 Washington University
 */
package org.nrg.xnat.tags;

import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.apache.log4j.Logger;
import org.dom4j.Document;
import org.dom4j.DocumentFactory;
import org.dom4j.Element;
import org.dom4j.Namespace;
import org.dom4j.QName;

/**
 * Servlet for tag label management
 * @author Kevin A. Archie <karchie@npg.wustl.edu>
 */
public final class TagLabelsServlet extends HttpServlet {
    private static final long serialVersionUID = 1L;
    private static final String JNDI_JDBC_DATASOURCE = "java:comp/env/jdbc/datasource";
    private static final String CONTENT_TYPE_LABELS = "application/x-xnat-metadata-labels+xml";
    private static final Namespace NAMESPACE_XE = new Namespace(null, "http://nrg.wustl.edu/xe");
    private static final QName XML_GET_ROOT = new QName("Labels", NAMESPACE_XE);
    private static final QName XML_GET_ELEM = new QName("Label", NAMESPACE_XE);
    private static final QName XML_DELETE_ROOT = new QName("Deleted", NAMESPACE_XE);
    private static final QName XML_DELETE_ELEM = new QName("Label", NAMESPACE_XE);

    private DataSource ds = null;


    public void init(final ServletConfig config) throws ServletException {
	super.init(config);
	try {
	    final InitialContext ic = new InitialContext();
	    ds = (DataSource)ic.lookup(JNDI_JDBC_DATASOURCE);
	} catch (NamingException e) {
	    throw new ServletException("Unable to obtain JDBC DataSource", e);
	}
    }


    private void doDBCheckpoint(final Connection c) {
	try {
	    final Statement s = c.createStatement();
	    try {
		s.execute("CHECKPOINT;");
	    } catch (SQLException e) {
		Logger.getLogger(TagLabelsServlet.class).warn(e);
	    } finally {
		try { s.close(); } catch (SQLException ignore) {}
	    }
	} catch (SQLException e) {
	    Logger.getLogger(TagLabelsServlet.class).warn("CHECKPOINT failed", e);
	}
    }


    /*
     * (non-Javadoc)
     * @see javax.servlet.http.HttpServlet#doPost(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
     */
    protected void doPost(final HttpServletRequest req, final HttpServletResponse res) throws IOException {
	// Extract tag labels from parameters and check each for validity.
	final Collection<String> invalidLabels = new ArrayList<String>();
	final Collection<TagLabel> tagLabels = TagLabel.makeLabels(req.getParameterNames(), invalidLabels);

	if (!invalidLabels.isEmpty()) {
	    res.sendError(HttpServletResponse.SC_BAD_REQUEST, "Invalid tag labels: " + invalidLabels);
	    return;
	}

	// Define tags in database
	try {
	    final Connection c = ds.getConnection();
	    try {
		TagOps.add(c, tagLabels);
		doDBCheckpoint(c);
	    } finally {
		try { c.close(); } catch (SQLException ignore) {}
	    }
	} catch (SQLException e) {
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;
	}
    }


    /*
     * (non-Javadoc)
     * @see javax.servlet.http.HttpServlet#doGet(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
     */
    protected void doGet(final HttpServletRequest req, final HttpServletResponse res) throws IOException {
	// Extract tag labels from parameters
	final Collection<String> invalidLabels = new ArrayList<String>();
	final Collection<TagLabel> tagLabels = TagLabel.makeLabels(req.getParameterMap().keySet(), invalidLabels);

	if (!invalidLabels.isEmpty()) {
	    res.sendError(HttpServletResponse.SC_BAD_REQUEST, "requested invalid tag labels: " + invalidLabels);
	    return;
	}

	final Document d = DocumentFactory.getInstance().createDocument();
	final Element root = d.addElement(XML_GET_ROOT);
	final Collection<TagLabel> labels;
	try {
	    final Connection c = ds.getConnection();
	    try {
		if (tagLabels.isEmpty()) {
		    labels = TagOps.getAll(c);
		} else {
		    labels = TagOps.getMatching(c, tagLabels);
		}
	    } finally {
		try { c.close(); } catch (SQLException ignore) {}
	    }
	} catch (SQLException e) {
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;
	}
	for (final TagLabel tl : labels) {
	    final Element labele = root.addElement(XML_GET_ELEM);
	    labele.setText(tl.toString());	   
	}

	res.setContentType(CONTENT_TYPE_LABELS);
	d.write(res.getWriter());
    }

    /*    protected void doPut(final HttpServletRequest req, final HttpServletResponse res) throws IOException {
	// TODO: Extract tag label and changes from parameters
	// TODO: Change tag labels in database
	res.sendError(HttpServletResponse.SC_NOT_IMPLEMENTED, "PUT operation not yet implemented");
    }
     */    

    protected void doDelete(final HttpServletRequest req, final HttpServletResponse res) throws IOException {
	// Extract tag labels from parameters
	final Collection<String> invalidLabels = new ArrayList<String>();
	final Collection<TagLabel> tagLabels = TagLabel.makeLabels(req.getParameterMap().keySet(), invalidLabels);

	if (!invalidLabels.isEmpty()) {
	    res.sendError(HttpServletResponse.SC_BAD_REQUEST, "requested invalid tag labels: " + invalidLabels);
	    return;
	}

	final Document d = DocumentFactory.getInstance().createDocument();
	final Element root = d.addElement(XML_DELETE_ROOT);

	final Collection<TagLabel> deleted;
	try {
	    final Connection c = ds.getConnection();
	    try {
		deleted = TagOps.delete(c, tagLabels);
		doDBCheckpoint(c);
	    } finally {
		try { c.close(); } catch (SQLException ignore) {}
	    }
	} catch (SQLException e) {
	    res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
	    return;
	}

	for (final TagLabel label : deleted) {
	    final Element e = root.addElement(XML_DELETE_ELEM);
	    e.setText(label.toString());
	}

	d.write(res.getWriter());
    }
}
